"""Embedded resource constants and helpers."""

import glob
import os


PACKAGE_PATH = os.path.realpath(os.path.dirname(__file__))

RESOURCES_PATH = os.path.join(PACKAGE_PATH, 'resources')

LAYOUTS_PATH = os.path.join(RESOURCES_PATH, 'layouts')

ASSETS_PATH = os.path.join(RESOURCES_PATH, 'assets')

TEMPLATE_SCREEN_IMAGE_PATHS = (
        os.path.join(ASSETS_PATH, '%d.%s'),
        os.path.join(ASSETS_PATH, 'reference', '%d.%s'),
)


def layout_paths(layout, recurse=True):
    path = os.path.join(LAYOUTS_PATH, layout)
    if recurse:
        path = os.path.join(path, '*.plp')
    return sorted(glob.glob(path))


def asset_path(index, fmt='raw'):
    for tmpl in TEMPLATE_SCREEN_IMAGE_PATHS:
        path = tmpl % (index, fmt)
        if os.path.exists(path):
            return path
    return None


def asset_file(index, fmt='raw'):
    return open(asset_path(index, fmt), 'rb')



from ambit.flags import FLAGS

import ambit.resources

import json
import sys

from typing import Any, Callable, Dict, List, Tuple


class Profile(object):
    title: str
    icon: int
    action_map: Dict
    orientation: int

    def __init__(self):
        self.title = "PALETTE"
        self.icon = Configuration.ICON_PALETTE
        self.action_map = {}
        self.media_map = {}
        self.joystick_map = {}
        self.color_map = {}
        self.key_map = {}
        self.layout = {}
        self.orientation = 0


class Configuration(object):
    ACTION_PROFILE_PREV = 'previousFunction'
    ACTION_PROFILE_NEXT = 'nextFunction'
    ACTION_PROFILE_SWITCH = 'switchFunctions'
    ACTION_TOGGLE_PREVIEW = 'togglePreview'
    ACTION_SET_COLOR_RED = 'setColorRed'
    ACTION_SET_COLOR_GREEN = 'setColorGreen'
    ACTION_SET_COLOR_BLUE = 'setColorBlue'
    ACTION_EXECUTE_COMMAND = 'executeCommand'
    ACTION_CYCLE_MAPPING = 'cycleMapping'

    ACTION_TEST_CYCLE = 'testCycle'
    ACTION_TEST_ACCUMULATE = 'testAccumulate'
    ACTION_TEST_DELTA = 'testDelta'
    ACTION_TEST_TRIGGER = 'testTrigger'

    # FIXME: add "behavior" keywords and mappings

    INPUT_PRESSED = 'pressed'
    INPUT_RELEASED = 'released'
    INPUT_LONG_PRESSED = 'long_pressed'
    INPUT_DOUBLE_PRESSED = 'double_pressed'
    INPUT_TRIPLE_PRESSED = 'triple_pressed'
    INPUT_ROTATION_RIGHT = 'rotation_right'
    INPUT_ROTATION_LEFT = 'rotation_left'
    INPUT_ROTATION = 'rotation'
    INPUT_PRESSED_ROTATION_RIGHT = 'pressed_rotation_right'
    INPUT_PRESSED_ROTATION_LEFT = 'pressed_rotation_left'
    INPUT_PRESSED_ROTATION = 'pressed_rotation'
    INPUT_SET = 'set'
    INPUT_MIXED = 'mixed'

    ICON_PALETTE = 0
    ICON_PHOTOSHOP = 1
    ICON_LIGHTROOM = 2
    ICON_AE = 3
    ICON_MIDI = 4
    ICON_KEYS = 5
    ICON_AI = 6
    ICON_JOYSTICK = 7
    ICON_COPYRIGHT = 8
    ICON_ID = 9
    ICON_PR = 10
    ICON_VIDEO = 11
    ICON_AU = 12
    ICON_CAPITAL_ONE = 13
    ICON_CHROME = 14
    ICON_SPOTIFY = 15
    ICON_MUSIC = 16
    ICON_VLC = 17
    ICON_CH = 18
    ICON_MEDIA = 19
    ICON_FLT = 20
    ICON_GEAR = 21
    #ICON_PALETTE = 22  # duplicate
    ICON_SHELL = 23
    ICON_HOME = 24
    ICON_BLANK = 25

    QT_KEY_MAP = {
        0x01000020: "PK_MOD_LSHIFT",
        0x01000021: "PK_MOD_LCTRL",
        0x01000022: "PK_MOD_LMETA",
        0x01000023: "PK_MOD_LALT",
    }

    HID_KEY_MAP = {
        "PK_MOD_LCTRL": 0x01,
        "PK_MOD_LSHIFT": 0x02,
        "PK_MOD_LALT": 0x04,
        "PK_MOD_LMETA": 0x08,
        "PK_MOD_RCTRL": 0x10,
        "PK_MOD_RSHIFT": 0x20,
        "PK_MOD_RALT": 0x40,
        "PK_MOD_RMETA": 0x80,
        "PK_NONE": 0x00,
        "PK_ERR_OVF": 0x01,
        "PK_A": 0x04,
        "PK_B": 0x05,
        "PK_C": 0x06,
        "PK_D": 0x07,
        "PK_E": 0x08,
        "PK_F": 0x09,
        "PK_G": 0x0a,
        "PK_H": 0x0b,
        "PK_I": 0x0c,
        "PK_J": 0x0d,
        "PK_K": 0x0e,
        "PK_L": 0x0f,
        "PK_M": 0x10,
        "PK_N": 0x11,
        "PK_O": 0x12,
        "PK_P": 0x13,
        "PK_Q": 0x14,
        "PK_R": 0x15,
        "PK_S": 0x16,
        "PK_T": 0x17,
        "PK_U": 0x18,
        "PK_V": 0x19,
        "PK_W": 0x1a,
        "PK_X": 0x1b,
        "PK_Y": 0x1c,
        "PK_Z": 0x1d,
        "PK_1": 0x1e,
        "PK_2": 0x1f,
        "PK_3": 0x20,
        "PK_4": 0x21,
        "PK_5": 0x22,
        "PK_6": 0x23,
        "PK_7": 0x24,
        "PK_8": 0x25,
        "PK_9": 0x26,
        "PK_0": 0x27,
        "PK_RETURN": 0x28,
        "PK_ESCAPE": 0x29,
        "PK_BACKSPACE": 0x2a,
        "PK_TAB": 0x2b,
        "PK_SPACE": 0x2c,
        "PK_MINUS": 0x2d,
        "PK_EQUAL": 0x2e,
        "PK_LEFTBRACE": 0x2f,
        "PK_RIGHTBRACE": 0x30,
        "PK_BACKSLASH": 0x31,
        "PK_HASHTILDE": 0x32,
        "PK_SEMICOLON": 0x33,
        "PK_APOSTROPHE": 0x34,
        "PK_GRAVE": 0x35,
        "PK_COMMA": 0x36,
        "PK_DOT": 0x37,
        "PK_SLASH": 0x38,
        "PK_CAPSLOCK": 0x39,
        "PK_F1": 0x3a,
        "PK_F2": 0x3b,
        "PK_F3": 0x3c,
        "PK_F4": 0x3d,
        "PK_F5": 0x3e,
        "PK_F6": 0x3f,
        "PK_F7": 0x40,
        "PK_F8": 0x41,
        "PK_F9": 0x42,
        "PK_F10": 0x43,
        "PK_F11": 0x44,
        "PK_F12": 0x45,
        "PK_SYSRQ": 0x46,
        "PK_SCROLLLOCK": 0x47,
        "PK_PAUSE": 0x48,
        "PK_INSERT": 0x49,
        "PK_HOME": 0x4a,
        "PK_PAGEUP": 0x4b,
        "PK_DELETE": 0x4c,
        "PK_END": 0x4d,
        "PK_PAGEDOWN": 0x4e,
        "PK_RIGHT": 0x4f,
        "PK_LEFT": 0x50,
        "PK_DOWN": 0x51,
        "PK_UP": 0x52,
        "PK_NUMLOCK": 0x53,
        "PK_KPSLASH": 0x54,
        "PK_KPASTERISK": 0x55,
        "PK_KPMINUS": 0x56,
        "PK_KPPLUS": 0x57,
        "PK_KPENTER": 0x58,
        "PK_KP1": 0x59,
        "PK_KP2": 0x5a,
        "PK_KP3": 0x5b,
        "PK_KP4": 0x5c,
        "PK_KP5": 0x5d,
        "PK_KP6": 0x5e,
        "PK_KP7": 0x5f,
        "PK_KP8": 0x60,
        "PK_KP9": 0x61,
        "PK_KP0": 0x62,
        "PK_KPDOT": 0x63,
        "PK_102ND": 0x64,
        "PK_COMPOSE": 0x65,
        "PK_POWER": 0x66,
        "PK_KPEQUAL": 0x67,
        "PK_F13": 0x68,
        "PK_F14": 0x69,
        "PK_F15": 0x6a,
        "PK_F16": 0x6b,
        "PK_F17": 0x6c,
        "PK_F18": 0x6d,
        "PK_F19": 0x6e,
        "PK_F20": 0x6f,
        "PK_F21": 0x70,
        "PK_F22": 0x71,
        "PK_F23": 0x72,
        "PK_F24": 0x73,
        "PK_OPEN": 0x74,
        "PK_HELP": 0x75,
        "PK_PROPS": 0x76,
        "PK_FRONT": 0x77,
        "PK_STOP": 0x78,
        "PK_AGAIN": 0x79,
        "PK_UNDO": 0x7a,
        "PK_CUT": 0x7b,
        "PK_COPY": 0x7c,
        "PK_PASTE": 0x7d,
        "PK_FIND": 0x7e,
        "PK_MUTE": 0x7f,
        "PK_VOLUMEUP": 0x80,
        "PK_VOLUMEDOWN": 0x81,
        "PK_KPCOMMA": 0x85,
        "PK_RO": 0x87,
        "PK_KATAKANAHIRAGANA": 0x88,
        "PK_YEN": 0x89,
        "PK_HENKAN": 0x8a,
        "PK_MUHENKAN": 0x8b,
        "PK_KPJPCOMMA": 0x8c,
        "PK_HANGEUL": 0x90,
        "PK_HANJA": 0x91,
        "PK_KATAKANA": 0x92,
        "PK_HIRAGANA": 0x93,
        "PK_ZENKAKUHANKAKU": 0x94,
        "PK_KPLEFTPAREN": 0xb6,
        "PK_KPRIGHTPAREN": 0xb7,
        "PK_LEFTCTRL": 0xe0,
        "PK_LEFTSHIFT": 0xe1,
        "PK_LEFTALT": 0xe2,
        "PK_LEFTWINDOWS": 0xe3,
        "PK_NOT_A_KEY": 0xe3,
        "PK_RIGHTCTRL": 0xe4,
        "PK_RIGHTSHIFT": 0xe5,
        "PK_RIGHTALT": 0xe6,
        "PK_RIGHTMETA": 0xe7,
        "PK_MEDIA_PLAYPAUSE": 0xe8,
        "PK_MEDIA_STOPCD": 0xe9,
        "PK_MEDIA_PREVIOUSSONG": 0xea,
        "PK_MEDIA_NEXTSONG": 0xeb,
        "PK_MEDIA_EJECTCD": 0xec,
        "PK_MEDIA_VOLUMEUP": 0xed,
        "PK_MEDIA_VOLUMEDOWN": 0xee,
        "PK_MEDIA_MUTE": 0xef,
        "PK_MEDIA_WWW": 0xf0,
        "PK_MEDIA_BACK": 0xf1,
        "PK_MEDIA_FORWARD": 0xf2,
        "PK_MEDIA_STOP": 0xf3,
        "PK_MEDIA_FIND": 0xf4,
        "PK_MEDIA_SCROLLUP": 0xf5,
        "PK_MEDIA_SCROLLDOWN": 0xf6,
        "PK_MEDIA_EDIT": 0xf7,
        "PK_MEDIA_SLEEP": 0xf8,
        "PK_MEDIA_COFFEE": 0xf9,
        "PK_MEDIA_REFRESH": 0xfa,
        "PK_MEDIA_CALC": 0xfb,
    }

    ICON_TYPE_MAP = {
        'Keyboard': ICON_KEYS,
        'Joystick': ICON_JOYSTICK,
        'MIDI': ICON_MIDI,
        'Palette': ICON_PALETTE,
        'Video': ICON_VIDEO,
        'Settings': ICON_GEAR,
        'Media': ICON_MEDIA,
        'Music': ICON_MUSIC,
        'Shell': ICON_SHELL,
        'Home': ICON_HOME,
    }

    profile_index: int
    profile: Profile
    profiles: List[Profile]

    def __init__(self, paths=None):
        self.profile_index = 0
        self.profile = Profile()
        self.profiles = []
        self.config_dicts = []
        if not paths:
            return
        for path in paths:
            with open(path, 'r') as f:
                config_bytes = f.read()
            try:
                config_dict = json.loads(config_bytes)
            except Exception as err:
                print('Error loading JSON from %s: %s' % (path, err))
                sys.exit(1)
            self.config_dicts.append(config_dict)
            profile = Profile()
            profile.title = config_dict['title']
            profile.icon = Configuration.ICON_TYPE_MAP[config_dict['tabType']]
            profile.action_map = config_dict['module_mappings'].get('actionMap', {})
            profile.media_map = config_dict['module_mappings'].get('mediaMap', {})
            profile.key_map = config_dict['module_mappings'].get('keyMap', {})
            profile.layout = config_dict.get('layout', {})
            profile.orientation = config_dict.get('orientation', 0)
            self.profiles.append(profile)
        self.switch(0)

    def switch(self, index):
        self.profile_index = index
        self.profile = self.profiles[index]
        print('[0] Switched to profile %d: %s' % (index, self.profile.title))

    def prev(self):
        index = self.profile_index - 1
        if index < 0:
            index = len(self.profiles) - 1
        self.switch(index)

    def next(self):
        index = self.profile_index + 1
        if index > len(self.profiles) - 1:
            index = 0
        self.switch(index)

    def component_hidkeys(self, uid):
        component_hidkeys = []
        for control in self.profile.key_map.get(uid, {}):
            bind_config = self.profile.key_map[uid].get(control, {})
            key = Configuration.HID_KEY_MAP[bind_config['virtual_code']]
            mod = 0
            for mc in bind_config['modifier_codes']:
                if type(mc) is int:
                    mc = Configuration.QT_KEY_MAP[mc]
                mod += Configuration.HID_KEY_MAP[mc]
            repeat = bind_config.get('dial_sensitivity', 1)
            if repeat != 1:
                repeat = 8
            component_hidkeys.append((control, key, mod, repeat))
        return component_hidkeys

    def components(self):
        """Return the list of configured components."""
        matchers = set()
        for uid_query in self.profile.action_map:
            matchers.add(uid_query)
        for uid_query in self.profile.media_map:
            matchers.add(uid_query)
        return sorted(list(matchers))

    def component_actions(self, uid):
        component_actions = []
        for input_type in self.profile.action_map.get(uid, {}):
            action_config = self.profile.action_map[uid].get(input_type, {})
            if 'action' not in action_config:
                continue
            action_name = action_config['action']
            component_actions.append((input_type, action_name, action_config))
        media_action = self.profile.media_map.get(uid, {})
        if media_action:
            action_name = media_action['key']
            input_type = Configuration.INPUT_MIXED
            component_actions.append((input_type, action_name, media_action))
        return component_actions

    def set_component_actions(self, uid, mapping):
        self.profile.action_map[uid] = mapping


def StandardConfiguration():
    config_paths = []
    for layout in FLAGS.layouts:
        layout_paths = ambit.resources.layout_paths(layout)
        if not layout_paths:
            print('[!] Unable to find layout %s in %s, exiting!' % (
                layout, ambit.resources.LAYOUTS_PATH))
            sys.exit(1)
        config_paths.extend(layout_paths)
    config_paths.extend(FLAGS.config_paths)
    config = Configuration(config_paths)
    # TODO: move the switch here out of Configuration.__init__()
    #config.switch(0)
    return config



import ctypes
import math
import os

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = 'hide'

import pygame

IMAGE_SIZE = (128, 128)

PALETTE_SIZE = 48


class ImageByte_bits(ctypes.BigEndianStructure):
    _fields_ = [
        ("a", ctypes.c_uint8, 4),
        ("b", ctypes.c_uint8, 4),
    ]


class ImageByte(ctypes.Union):
    _anonymous_ = ("bit",)
    _fields_ = [ ("bit", ImageByte_bits), ("asByte", ctypes.c_uint8) ]


def load_image(path, image_size=IMAGE_SIZE, palette_size=PALETTE_SIZE):
    with open(path, 'rb') as f:
        return parse_image(f, image_size, palette_size)


def parse_image(fd, image_size=IMAGE_SIZE, palette_size=PALETTE_SIZE):
    # Prepare a one-dimensional sequence of pixels
    pixels = []
    colors = []

    # read the palette
    for _ in range(int(palette_size / 3)):
        r = ord(fd.read(1))
        g = ord(fd.read(1))
        b = ord(fd.read(1))
        colors.append((b, g, r))

    # read the pixels
    body_size = int(image_size[0] / 2) * image_size[1]
    for _ in range(body_size):
        d = fd.read(1)

        k = ImageByte()
        k.asByte = ord(d)
        pixels.append(k.a)
        pixels.append(k.b)

    # not sure what this is?
    fd.read(4)
    fd.read(64)

    # extra bytes
    extra = fd.read()

    if len(extra):
        print('[!] WARNING: image input had extra bytes %d' % len(extra))

    return colors, pixels


def draw_pixels(surface, pixels, image_size=IMAGE_SIZE):
    # Draw the result to the surface
    i = 0
    for y in range(image_size[0], 0, -1):
        for x in range(image_size[1]):
            px = pixels[i]
            surface.set_at((x, y), px)
            i += 1


def surface(path, image_size=IMAGE_SIZE, palette_size=PALETTE_SIZE, surface_size=IMAGE_SIZE):
    surface = pygame.Surface(image_size, depth=8)
    colors, pixels = load_image(path, image_size, palette_size)
    surface.set_palette(colors)
    draw_pixels(surface, pixels, image_size)
    return pygame.transform.scale(surface, surface_size)


def load_standard_image(path, palette_size, image_size=IMAGE_SIZE):
    pygame.display.init()

    surface = pygame.image.load(path).convert(24)
    surface = pygame.transform.scale(surface, image_size)

    palette = []
    pixels = []
    seen_colors = {}
    palette_map = {}

    # first, grab all of the pixels in the image
    # and build a list of most common pixel colors
    for y in range(image_size[1]):
        for x in range(image_size[0] - 1, -1, -1):
            color = surface.get_at((x, y))
            c = surface.map_rgb(color)
            pixels.append(c)
            if not c in seen_colors: seen_colors[c] = 0
            seen_colors[c] += 1

    top_colors = [0] + sorted(seen_colors, key=seen_colors.get, reverse=True)
    num_colors = int(palette_size / 3)
    for i, c in enumerate(top_colors[:num_colors]):
        palette_map[c] = i
        color = pygame.Color(c)
        _, r, g, b = color
        palette.append(b)
        palette.append(g)
        palette.append(r)
    for i in range(16 - int(len(palette) / 3)):
        palette.extend((0, 0, 0))

    return palette, palette_map, pixels


def convert_image(src_path, dst_fd):
    palette, palette_map, pixels = load_standard_image(src_path, palette_size=PALETTE_SIZE)
    data = make_image_bytes(palette, palette_map, pixels)
    dst_fd.write(data)


def make_image_bytes(palette, palette_map, pixels):
    data = bytearray()
    mapped_pixels = []

    # add the palette mapped pixels to the body
    k = ImageByte()
    cur_nibble = 0
    for px in pixels:
        px_color = pygame.Color(px)
        color_diffs = []
        for c in palette_map:
            color = pygame.Color(c)
            diff = math.sqrt(abs(px_color.r - color.r)**2 + abs(px_color.g - color.g)**2 + abs(px_color.b - color.b)**2)
            color_diffs.append((diff, c))
        palette_index = palette_map[min(color_diffs)[1]]
        if cur_nibble == 0:
            k.b = palette_index
            cur_nibble = 1
            continue
        if cur_nibble == 1:
            k.a = palette_index
            mapped_pixels.append(k.asByte)
            cur_nibble = 0

    data.extend(palette)
    data.extend(reversed(mapped_pixels))

    if len(data) != 8240:
        print('[!] WARNING: image bytes output had %d extra bytes' % 8240 - len(data))

    return data



"""Fake device for testing."""

import ambit

import io
import queue
import random
import time
import usb


LAYOUT_DEFAULT_EXPERTKIT = {
        'u': '00000', 'i': 1, 't': ambit.Component.KIND_BASE, 'c': [
            None,
            {'u': '  49d', 'i': 2, 't': ambit.Component.KIND_DIAL, 'c': [
                {'u': '  4FH', 'i': 3, 't': ambit.Component.KIND_DIAL, 'c': [
                    {'u': '  4lf', 'i': 8, 't': ambit.Component.KIND_BUTTON, 'c': [None, None, None]},
                    {'u': '  4}=', 'i': 4, 't': ambit.Component.KIND_SLIDER, 'c': [None, None, None, None, None]},
                    None,
                ]},
                None,
                {'u': '  4A0', 'i': 5, 't': ambit.Component.KIND_DIAL, 'c': [
                    None,
                    None,
                    {'u': '  4ls', 'i': 6, 't': ambit.Component.KIND_BUTTON, 'c': [
                        {'u': '  4}M', 'i': 7, 't': ambit.Component.KIND_SLIDER, 'c': [None, None, None, None, None]},
                        None,
                        None
                    ]},
                ]},
            ]},
            None,
        ]}

LAYOUT_DEFAULT_PROKIT = {
        'u': '00000', 'i': 1, 't': 0, 'c': [
            None,
            {'u': '  C4@', 'i': 2, 't': 2, 'c': [
                {'u': '  D5o', 'i': 3, 't': 2, 'c': [
                    {'u': '  B$;', 'i': 4, 't': 1, 'c': [
                        None,
                        {'u': '  8S$', 'i': 5, 't': 1, 'c': [None, None, None]},
                        None,
                    ]},
                    {'u': '  Ak~', 'i': 11, 't': 3, 'c': [
                        None,
                        {'u': '  6om', 'i': 12, 't': 2, 'c': [None, None, None]},
                        None,
                        None,
                        None,
                    ]},
                    {'u': '  8Uq', 'i': 13, 't': 3, 'c': [None, None, None, None, None]},
                ]},
                {'u': '  6pZ', 'i': 6, 't': 2, 'c': [None, None, None]},
                {'u': '  6RD', 'i': 7, 't': 2, 'c': [
                    None,
                    None,
                    {'u': '  8a4', 'i': 8, 't': 1, 'c': [
                        {'u': '  8W(', 'i': 9, 't': 3, 'c': [
                            None,
                            {'u': '  8V$', 'i': 10, 't': 3, 'c': [None, None, None, None, None]},
                            None,
                            None,
                            None,
                        ]},
                        {'u': '  8R+', 'i': 14, 't': 1, 'c': [
                            {'u': '  6J@', 'i': 15, 't': 2, 'c': [None, None, None]},
                            None,
                            None
                        ]},
                        None,
                    ]},
                ]},
            ]},
            None,
        ]}

LAYOUT_BASE_ONLY = {
        'u': '00000', 'i': 1, 't': ambit.Component.KIND_BASE,
        'c': [None, None, None]}

LAYOUT_DEFAULT = LAYOUT_DEFAULT_PROKIT

def make_quad(cur):
    index1 = len(cur) + 1
    index2 = index1 + 1
    index3 = index2 + 1
    index4 = index3 + 1
    uid1 = '  4%.2x' % index1
    uid2 = '  4%.2x' % index2
    uid3 = '  4%.2x' % index3
    uid4 = '  4%.2x' % index4
    kind1 = random.choice([
        ambit.Component.KIND_BUTTON,
        ambit.Component.KIND_DIAL,
    ])
    kind2 = random.choice([
        ambit.Component.KIND_BUTTON,
        ambit.Component.KIND_DIAL,
    ])
    kind3 = random.choice([
        ambit.Component.KIND_BUTTON,
        ambit.Component.KIND_DIAL,
    ])
    kind4 = random.choice([
        ambit.Component.KIND_BUTTON,
        ambit.Component.KIND_DIAL,
    ])
    quad = {
        'u': uid1, 'i': index1, 't': kind1, 'c': [
            {'u': uid2, 'i': index2, 't': kind2, 'c': [None, None, None]},
            {'u': uid3, 'i': index3, 't': kind3, 'c': [
                {'u': uid4, 'i': index4, 't': kind4, 'c': [None, None, None]},
                None,
                None,
            ]},
            None,
        ]}
    cur.extend([1,1,1,1])
    return quad


def make_slider_quad(cur):
    index1 = len(cur) + 1
    index2 = index1 + 1
    uid1 = '  4%.2x' % index1
    uid2 = '  4%.2x' % index2
    kind = ambit.Component.KIND_SLIDER
    quad = {
        'u': uid1, 'i': index1, 't': kind, 'c': [
            None,
            None,
            None,
            {'u': uid2, 'i': index2, 't': kind, 'c': [None, None, None, None, None]},
            None,
        ]}
    cur.extend([1,1])
    return quad


def make_component_layout():
    cur = []
    layout = make_quad(cur)
    layout['u'] = '00000'
    layout['i'] = 1
    layout['t'] = ambit.Component.KIND_BASE

    # attach right
    layout['c'][1]['c'][0]['c'][1] = make_quad(cur)
    # attach below
    layout['c'][1]['c'][1] = make_quad(cur)
    # attach left
    layout['c'][2] = make_quad(cur)

    # attach below-right
    layout['c'][1]['c'][0]['c'][1]['c'][2] = make_slider_quad(cur)
    # attach below-left
    layout['c'][2]['c'][1]['c'][0]['c'][1] = make_slider_quad(cur)

    # attach below-right x 2
    #layout['c'][1]['c'][0]['c'][1]['c'][2]['c'][3]['c'][1] = make_slider_quad(cur)
    # attach below-left x 2
    #layout['c'][2]['c'][1]['c'][0]['c'][1]['c'][4] = make_slider_quad(cur)

    return layout


LAYOUT_RANDOM = make_component_layout()


# NOTE: it is best to rely on as little functionality from the main
# ambit library as possible here.

class Handle:
    PHASE_CONTROL_0 = 'CONTROL_0'
    PHASE_CONTROL_1 = 'CONTROL_1'
    PHASE_CONTROL_2 = 'CONTROL_2'
    PHASE_CONTROL_3 = 'CONTROL_3'
    PHASE_CONTROL_4 = 'CONTROL_4'
    PHASE_CONTROL_5 = 'CONTROL_5'
    PHASE_BULK = 'BULK'
    PHASE_INPUT = 'INPUT'

    # TODO: scale these with number of attached components
    # based on analysis in docs/PERFORMANCE.md
    # TODO: consider merging this with the consts in ambit.Controller
    DELAY_SCREEN_STRING_SECONDS = 1 / 20
    DELAY_LED_SECONDS = 1 / 60

    # TODO: reduce the max queue depth once we have a
    # reasonable way to kick off test inputs after we are
    # sure the Controller read thread is ready.
    WRITE_QUEUE_DEPTH = 32

    def __init__(self, layout):
        self._write_queue = queue.Queue(Handle.WRITE_QUEUE_DEPTH)
        self._layout = layout
        self._last_check = 0
        self._dropped_writes = 0
        self.messages = {}
        self.leds = {}
        self.phase = None
        self.version = '1.3.1'
        self.screen_display = None
        self.screen_string = None
        self.screen_images = {}
        self._set_phase(Handle.PHASE_CONTROL_0)

    def _process_input(self, message, data):
        for message_type in message:
            if message_type not in self.messages:
                self.messages[message_type] = []
            self.messages[message_type].append(message[message_type])
        if 'start' in message:
            self._process_start(message['start'])
        if 'stop' in message:
            self._process_stop()
        if 'led' in message:
            self._process_led(message['led'])
        if 'check' in message:
            self._process_check(message['check'])
        if 'screen_display' in message:
            self.screen_display = message['screen_display']
        if 'screen_string' in message:
            time.sleep(Handle.DELAY_SCREEN_STRING_SECONDS)
            self.screen_string = message['screen_string']
        if 'screen_write' in message:
            self._process_screen_write(message['screen_write'], data)
        if 'send_version' in message:
            self._process_send_version()

    def _process_check(self, check):
        #print('[F] Processing check message')
        self._last_check = time.time()

    def _process_send_version(self):
        self.write_messages([{
            'version_core': self.version,
            'version_screen': 0,
        }])

    def _process_screen_write(self, index, image_bytes):
        fd = io.BytesIO(image_bytes)
        self.screen_images[index] = image_bytes

    def _process_start(self, start):
        #print('[F] Processing start message')
        self._write_layout()
        if self.phase == Handle.PHASE_BULK:
            self._set_phase(Handle.PHASE_INPUT)

    def _process_stop(self):
        self._set_phase(Handle.PHASE_BULK)

    def _process_led(self, leds):
        #print('[F] Processing led message', leds)
        time.sleep(Handle.DELAY_LED_SECONDS)
        for led in leds:
            self.leds[led['i']] = (led['r'], led['g'], led['b'])

    def _write_layout(self):
        #print('[F] Sending layout')
        self.write_messages([
            {'l': self._layout},
        ])

    def write_messages(self, messages):
        data = ambit.message_encode(messages)
        self._write(data)

    def _write(self, data):
        if self._write_queue.full():
            for _ in range(int(self._write_queue.qsize() / 8)):
                self._write_queue.get()
                self._write_queue.task_done()
                self._dropped_writes += 1
                #print('DROPPED:', self._dropped_writes)
        self._write_queue.put(data)

    def _set_phase(self, phase):
        if phase == self.phase:
            return
        self.phase = phase
        if ambit.FLAGS.debug:
            print('[F] Device Phase Change:', self.phase)

    def bulkWrite(self, ep, data, size, timeout=0):
        messages, extra_data = ambit.message_decode(data)
        if ambit.FLAGS.debug:
            print('[F] Handle.bulkWrite()', ep, messages, size, timeout)
        for message in messages:
            self._process_input(message, extra_data)
        return len(data)

    def bulkRead(self, ep, size, timeout=0):
        data = memoryview(bytes())
        start_time = time.time()
        while (not timeout) or (time.time() - start_time) * 1000 < timeout:
            try:
                data = self._write_queue.get(timeout=0.01)
                break
            except queue.Empty:
                pass
        if ambit.FLAGS.debug:
            print('[F] Handle.bulkRead()', ep, size, timeout, '=>', ambit.message_decode(data))
        if not data:
            raise usb.USBError('no data to read')
        return data

    def controlMsg(self, ep, ctrl, size):
        if ambit.FLAGS.debug:
            print('[F] Handle.controlMsg()', ep, ctrl, size)
        if self.phase == Handle.PHASE_CONTROL_0:
            self._set_phase(Handle.PHASE_CONTROL_1)
            return
        if self.phase == Handle.PHASE_CONTROL_1:
            self._set_phase(Handle.PHASE_CONTROL_2)
            return
        if self.phase == Handle.PHASE_CONTROL_2:
            self._set_phase(Handle.PHASE_CONTROL_3)
            return
        if self.phase == Handle.PHASE_CONTROL_3:
            self._set_phase(Handle.PHASE_CONTROL_4)
            return
        if self.phase == Handle.PHASE_CONTROL_4:
            self._set_phase(Handle.PHASE_CONTROL_5)
            return
        if self.phase == Handle.PHASE_CONTROL_5:
            self._set_phase(Handle.PHASE_BULK)
            return

    def reset(self):
        if ambit.FLAGS.verbose:
            print('[F] Handle.reset()')

    def releaseInterface(self):
        if ambit.FLAGS.verbose:
            print('[F] Handle.releaseInterface()')


class Device:
    def __init__(self, device_id, interface_id):
        if ambit.FLAGS.verbose:
            print('[F] Device.__init__()', device_id, interface_id)
        self.device_id = device_id
        self.interface_id = interface_id
        self._layout = {}
        self._version = ''
        self.values = {}
        self.kinds = {}
        self.handle = None

    def components_connected(self, layout=LAYOUT_DEFAULT):
        if ambit.FLAGS.debug:
            print('[F] Layout set to', layout)
        self._layout = layout
        self.values = {}
        self.initialize_values(self._layout)

    # TODO: consolidate this with components_connected?
    def layout_changed(self, layout):
        self.components_connected(layout)
        self.handle._layout = layout
        self.handle._write_layout()

    # NOTE: Has no effect if called after open()
    def set_version_core(self, version):
        self._version = version

    def initialize_values(self, component):
        if not component:
            return
        index = component['i']
        self.values[index] = [0,0,0,0,0,0,0,0]
        self.kinds[index] = component['t']
        for c in component['c']:
            self.initialize_values(c)

    def open(self):
        if ambit.FLAGS.verbose:
            print('[F] Device.open()')
        self.handle = Handle(self._layout)
        if self._version:
            self.handle.version = self._version
        return self.handle

    def input_pressed(self, index):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] not in (
                ambit.Component.KIND_DIAL,
                ambit.Component.KIND_BUTTON):
            return
        values = self.values[index]
        values[0] = 1
        values[1] = 0
        values[2] = 0
        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])

    def input_released(self, index):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] not in (
                ambit.Component.KIND_DIAL,
                ambit.Component.KIND_BUTTON):
            return
        values = self.values[index]
        values[0] = 0
        values[1] = 0
        values[2] = 0
        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])

    def input_slide_up(self, index, amount=8):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] != ambit.Component.KIND_SLIDER:
            return
        values = self.values[index]

        # With firmware versions <= 1.3.1, the maximum slider
        # value is 254 instead of 255.
        max_slider_value = 255
        if self.handle.version < ambit.Controller.FIRMWARE_VERSION_SLIDER_FIXED:
            max_slider_value = 254

        if values[0] == max_slider_value:
            return
        values[0] += amount
        if values[0] > max_slider_value:
            values[0] = max_slider_value

        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])

    def input_slide_down(self, index, amount=8):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] != ambit.Component.KIND_SLIDER:
            return
        values = self.values[index]
        if values[0] == 0:
            return
        values[0] -= amount
        if values[0] < 0:
            values[0] = 0
        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])

    def input_rotation_left(self, index, amount=8):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] != ambit.Component.KIND_DIAL:
            return
        values = self.values[index]
        values[1] = amount
        values[2] = 0
        values[3] -= amount
        if values[3] < 0:
            values[3] = 0
        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])

    def input_rotation_right(self, index, amount=8):
        if self.handle.phase != Handle.PHASE_INPUT:
            return

        if self.kinds[index] != ambit.Component.KIND_DIAL:
            return
        values = self.values[index]
        values[1] = 0
        values[2] = amount
        values[3] += amount
        if values[3] > 255:
            values[3] = 255
        self.values[index] = values
        self.handle.write_messages([
                {'in': [{'i': index, 'v': values}]},
        ])



import os

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

import ambit
import ambit.image
import ambit.resources
import pygame
import threading
import time

DISPLAY_SIZE = (1024, 576)
DISPLAY_SCALING_FACTOR = 1  # Typical display is 1, HiDPI is 2

# Times per second to check for events.
TICKS_PER_SECOND = 60

# Times per second to redraw component state.
REDRAWS_PER_SECOND = 30

# Times per second to update layout.
LAYOUT_UPDATES_PER_SECOND = 5

# Display size scaled to DPI setting.
SURFACE_SIZE = (
        DISPLAY_SIZE[0] * DISPLAY_SCALING_FACTOR,
        DISPLAY_SIZE[1] * DISPLAY_SCALING_FACTOR)

COMPONENT_COLOR = (164, 164, 164)
COMPONENT_SELECTED_COLOR = (164, 164, 255)

COMPONENT_MARGIN = 4 * DISPLAY_SCALING_FACTOR


def initial_pos_and_slot_size(surface_size, min_slot_x, min_slot_y, max_slot_x, max_slot_y):
    count_slots_x = abs(min_slot_x - max_slot_x) + 1
    count_slots_y = abs(min_slot_y - max_slot_y) + 1
    margin_x = count_slots_x * COMPONENT_MARGIN + (COMPONENT_MARGIN * 3)
    margin_y = count_slots_y * COMPONENT_MARGIN + (COMPONENT_MARGIN * 3)
    slot_width = int((surface_size[0] - margin_x) / count_slots_x)
    slot_height = int((surface_size[1] - margin_y) / count_slots_y)
    slot_size = min([slot_width, slot_height])
    initial_pos_x = (slot_width * -min_slot_x) + int(margin_x / 2)
    initial_pos_y = (slot_height * max_slot_y) + int(margin_y / 2)
    return (initial_pos_x, initial_pos_y), slot_size


# FIXME: could do more caching of calculations here
class ComponentRenderer(object):
    def __init__(self, initial_pos, slot_size, component, screen_display_surface_cache, color=COMPONENT_COLOR):
        self.component = component
        self.slot = component.slot
        self.orientation = component.orientation
        self.color = color
        self.initial_pos = initial_pos
        self.slot_size = slot_size

        self.screen_display_surface_cache = screen_display_surface_cache
        self.cached_screen_display = None

        self.wide = False
        if component.kind == ambit.Component.KIND_SLIDER:
            self.wide = True

        self.width = self.height = slot_size
        if self.wide:
            self.width += slot_size + COMPONENT_MARGIN

        self.rotate_and_place()

        self.connector_size = int(slot_size / 32)
        self.font_size = int(slot_size / 16)
        self.screen_display_size = (
            int(slot_size / 1.5),
            int(slot_size / 1.5))
        self.led_thickness = int(slot_size / 32)
        self.id_inset = self.led_thickness + COMPONENT_MARGIN * 2.2

        self.component_rect = pygame.Rect(self.x, self.y, self.w, self.h)
        self.led_rect = pygame.Rect(
            self.x + self.led_thickness, self.y + self.led_thickness,
            self.w - (self.led_thickness * 2), self.h - (self.led_thickness * 2))
        self.font = pygame.font.Font(pygame.font.get_default_font(), self.font_size)
        self.display_font = pygame.font.Font(
                pygame.font.get_default_font(), self.font_size + 4)

    def sync(self):
        if self.component.index == 1:
            self.sync_screen_display()

    def sync_screen_display(self):
        if self.cached_screen_display == self.component.screen_display:
            return

        self.screen_display_surface = None

        if self.component.screen_display in self.screen_display_surface_cache:
            self.screen_display_surface = self.screen_display_surface_cache[self.component.screen_display]

        if not self.screen_display_surface:
            asset_path = ambit.resources.asset_path(self.component.screen_display)
            try:
                self.screen_display_surface = ambit.image.surface(
                        asset_path, surface_size=self.screen_display_size)
            except:
                self.screen_display_surface = None

        if not self.screen_display_surface:
            self.screen_display_surface = pygame.Surface(ambit.image.IMAGE_SIZE)
            self.screen_display_surface.fill((0,0,0))

        self.screen_display_surface_cache[self.component.screen_display] = self.screen_display_surface
        self.cached_screen_display = self.component.screen_display

    def slot_coordinates(self):
        slotx, sloty = self.slot
        return (self.initial_pos[0] + (self.slot_size + COMPONENT_MARGIN) * slotx,
                self.initial_pos[1] - (self.slot_size + COMPONENT_MARGIN) * sloty)

    def orient(self):
        x, y = self.slot_coordinates()
        if self.orientation == 90:
            return x, y, self.height, self.width
        elif self.orientation == 270:
            return x, y - (self.width - self.slot_size), self.height, self.width
        elif self.orientation == 180:
            return x - (self.width - self.slot_size), y, self.width, self.height
        return x, y, self.width, self.height

    # FIXME: prepare a surface with leds, ports, etc.
    # in 0deg orientation and then rotate the entire
    # surface and blit? Text will still need to be
    # rendered after rotation.
    def root_port_position(self):
        div = 2
        if self.wide:
            div = 4
        if self.orientation == 0:
            return self.x + int(self.w / div), self.y
        if self.orientation == 90:
            return self.x + self.w, self.y + int(self.h / div)
        if self.orientation == 180:
            if self.wide:
                div = 1.33
            return self.x + int(self.w / div), self.y + self.h
        if self.orientation == 270:
            if self.wide:
                div = 1.33
            return self.x, self.y + int(self.h / div)

    def rotate_and_place(self):
        self.x, self.y, self.w, self.h = self.orient()
        self.root_port_pos = self.root_port_position()

    def draw(self, screen):
        self.draw_body(screen)
        self.draw_ports(screen)
        self.draw_id(screen)
        self.draw_slot(screen)
        self.draw_behavior(screen)
        self.draw_led(screen)
        self.draw_display(screen)
        self.draw_values(screen)

    def draw_body(self, screen):
        pygame.draw.rect(screen, self.color, self.component_rect)

    def draw_ports(self, screen):
        self.draw_root_port(screen)

    def draw_root_port(self, screen):
        pygame.draw.circle(screen, (255, 0, 0, 128),
                self.root_port_pos, self.connector_size)

    def draw_led(self, screen):
        pygame.draw.rect(screen, self.component.led,
                self.led_rect, self.led_thickness)

    def draw_id(self, screen):
        text = '[%s] %s (%s)' % (
                self.component.uid, self.component.index, self.component.kind_name())
        surface = self.font.render(text, True, (255, 255, 255))
        screen.blit(surface, (self.x + self.id_inset, self.y + self.id_inset))

    def draw_slot(self, screen):
        text = str(self.component.slot)
        surface = self.font.render(text, True, (255, 255, 255))
        screen.blit(surface, (self.x + self.id_inset, self.y + self.h - self.id_inset - surface.get_height()))

    def draw_behavior(self, screen):
        text = ''
        for input_type in self.component.callbacks:
            invocation = self.component.callbacks[input_type]
            if invocation.behavior:
                text = invocation.behavior.behavior
                break
        surface = self.font.render(text, True, (255, 255, 255))
        screen.blit(surface, (self.x + self.w - self.id_inset - surface.get_width(),
            self.y + self.h - self.id_inset - surface.get_height()))

    def draw_display(self, screen):
        if self.component.kind != ambit.Component.KIND_BASE:
            return

        isw, ish = self.screen_display_surface.get_size()
        isx = self.x + int(self.w / 2) - int(isw / 2)
        isy = self.y + int(self.h / 2) - int(ish / 2)
        screen.blit(self.screen_display_surface, (isx, isy))

        text = str(self.component.screen_string)
        fsurface = self.display_font.render(text, True, (255, 255, 255))
        fsw, fsh = fsurface.get_size()
        screen.blit(fsurface,
                (self.x + int(self.w / 2) - int(fsw / 2),
                isy + ish - int(fsh * 1.5)))

    def draw_values(self, screen):
        if self.component.kind == ambit.Component.KIND_BASE:
            return
        text = str(self.component.values)
        surface = self.font.render(text, True, (255, 255, 255))
        sw, sh = surface.get_size()
        screen.blit(surface,
                (self.x + int(self.w / 2) - int(sw / 2),
                self.y + int(self.h / 2) - int(sh / 2)))


class Display(object):
    def __init__(self, ctrl, close_on_exit=False, shutdown_event=None,
            pygame_event_callbacks=None):
        self.screen = None
        self.surface = None
        self.surface_size = SURFACE_SIZE
        self.ctrl = ctrl
        self.close_on_exit = close_on_exit
        self.selected_component = 0
        self.pygame_event_callbacks = pygame_event_callbacks
        self.screen_display_surface_cache = {}
        self.cached_extrema = None
        if shutdown_event:
            self.shutdown_event = shutdown_event
        else:
            self.shutdown_event = threading.Event()
        self.thread = threading.Thread(target=self.worker)

    def open(self):
        pygame.display.init()
        pygame.font.init()
        pygame.key.set_repeat(500, 100)
        self.update_display()
        self.layout_changed()

    def update_display(self):
        self.screen = pygame.display.set_mode(self.surface_size, pygame.RESIZABLE)
        self.surface = pygame.Surface(self.surface_size)
        self.screen.fill((0, 0, 0))
        pygame.display.flip()
        self.cached_extrema = None
        self.layout_changed()

    def prepare_sprites(self, initial_pos, slot_size, root):
        sprites = []
        sp = ComponentRenderer(
                initial_pos, slot_size, root,
                self.screen_display_surface_cache)
        sprites.append(sp)
        for port in root.children:
            component = root.children[port]
            if not component:
                continue
            sprites.extend(self.prepare_sprites(initial_pos, slot_size, component))
        return sprites

    def layout_changed(self):
        extrema = self.ctrl.layout.extrema()
        if self.cached_extrema == extrema:
            return
        min_x, min_y, max_x, max_y = extrema
        initial_pos, slot_size = initial_pos_and_slot_size(
                self.surface_size, min_x, min_y, max_x, max_y)
        root = self.ctrl.layout.find_component(1)
        self.sprites = self.prepare_sprites(initial_pos, slot_size, root)
        self.screen.fill((0, 0, 0))
        self.cached_extrema = min_x, min_y, max_x, max_y
        self.screen_display_surface_cache = {}
        # TODO: draw the underlying grid

    def draw(self):
        for sp in self.sprites:
            sp.sync()
            if sp.component.index == self.selected_component:
                sp.color = COMPONENT_SELECTED_COLOR
            else:
                sp.color = COMPONENT_COLOR
            sp.draw(self.screen)

    def shutdown(self):
        print('[R] Shutting down at user request.')
        self.shutdown_event.set()
        if self.close_on_exit:
            self.ctrl.close()

    def tick(self):
        # TODO: only redraw if layout *or* component state changed.
        if self.tick_count % int(TICKS_PER_SECOND / LAYOUT_UPDATES_PER_SECOND) == 0:
            pygame.display.flip()
            self.layout_changed()
            self.draw()
            pygame.display.flip()
        if self.tick_count % int(TICKS_PER_SECOND / REDRAWS_PER_SECOND) == 0:
            pygame.display.flip()
            self.draw()
            pygame.display.flip()

        event = pygame.event.poll()
        if event.type == pygame.NOEVENT:
            return

        if event.type == pygame.VIDEORESIZE:
            self.surface_size = event.size
            self.update_display()
        elif event.type == pygame.QUIT:
            self.shutdown()
        elif event.type == pygame.KEYDOWN and event.key in (
                pygame.K_q, pygame.K_ESCAPE):
            self.shutdown()
        elif event.type == pygame.KEYUP and event.key == pygame.K_PAGEUP:
            self.selected_component = (self.selected_component - 1) % (len(self.ctrl.layout.connected()) + 1)
        elif event.type == pygame.KEYUP and event.key == pygame.K_PAGEDOWN:
            self.selected_component = (self.selected_component + 1) % (len(self.ctrl.layout.connected()) + 1)
        elif event.type == pygame.KEYUP and event.key == pygame.K_TAB:
            self.ctrl.rotate(90)
        elif event.type == pygame.KEYDOWN:
            k_int_map = {
                    pygame.K_0: 0,
                    pygame.K_1: 1,
                    pygame.K_2: 2,
                    pygame.K_3: 3,
                    pygame.K_4: 4,
                    pygame.K_5: 5,
                    pygame.K_6: 6,
                    pygame.K_7: 7,
                    pygame.K_8: 8,
                    pygame.K_9: 9,
            }
            if event.key in k_int_map:
                self.selected_component = k_int_map[event.key]

        if self.pygame_event_callbacks:
            for callback in self.pygame_event_callbacks:
                callback(event)

    def run(self):
        self.open()
        self.thread.start()

    def worker(self):
        self.tick_count = 0
        while not self.shutdown_event.is_set():
            self.tick()
            self.tick_count += 1
            time.sleep(1 / TICKS_PER_SECOND)



from ambit.configuration import Configuration
from ambit.flags import FLAGS

import ambit.coordinates

import copy

from typing import Any, Callable, Dict, List, Tuple


def sorted_rowwise(components):
    return sorted(components, key=lambda x: (-x.slot[1], x.slot[0]))


class ComponentInvocation(object):
    input_type: int
    action_name: str
    behavior: Any
    function: Callable
    vindex: int
    action_config: Any
    persistent_state: Dict[str, Any]
    threshold: int
    data: Any
    cumulative_change: int

    def __init__(self, input_type, action_name, behavior, callback, vindex, action_config, persistent_state):
        self.input_type = input_type
        self.action_name = action_name
        self.behavior = behavior
        self.function = callback
        self.vindex = vindex
        self.action_config = action_config
        self.persistent_state = persistent_state

        if not 'stored_value' in self.persistent_state:
            self.persistent_state['stored_value'] = self.behavior.default

        self.threshold = self.behavior.threshold
        self.data = behavior.data

        self.cumulative_change = 0

    def callback_value(self, value):
        callback_value = value

        change = self.cumulative_change

        if self.behavior.behavior == ComponentBehavior.CYCLE:
            change = 1

        if self.behavior.invert:
            change = -change

        if self.behavior.behavior == ComponentBehavior.TRIGGER:
            callback_value = value
            if self.input_type == Configuration.INPUT_SET and self.behavior.limits:
                step = 255 / (self.behavior.limits[1] - self.behavior.limits[0])
                callback_value = self.behavior.limits[0] + value / step

        if self.behavior.behavior == ComponentBehavior.CYCLE:
            callback_value = self.persistent_state['stored_value'] + change
            if self.input_type == Configuration.INPUT_SET and self.behavior.limits:
                step = 255 / (self.behavior.limits[1] - self.behavior.limits[0])
                callback_value = int(value / step)

        if self.behavior.behavior == ComponentBehavior.DELTA:
            callback_value = self.cumulative_change
            if self.behavior.invert:
                callback_value = -callback_value

        if self.behavior.value is not None:
            callback_value = self.behavior.value

        if self.behavior.behavior == ComponentBehavior.ACCUMULATE:
            change = self.threshold
            if self.behavior.value:
                change = self.behavior.value
            if self.behavior.invert:
                change = -change
            callback_value = self.persistent_state['stored_value'] + change

        if self.behavior.loop:
            if callback_value < self.behavior.limits[0]:
                callback_value = self.behavior.limits[1]
            if callback_value > self.behavior.limits[1]:
                callback_value = self.behavior.limits[0]
        elif self.behavior.outside(callback_value):
            if callback_value < self.behavior.limits[0]:
                callback_value = self.behavior.limits[0]
            if callback_value > self.behavior.limits[1]:
                callback_value = self.behavior.limits[1]

        if self.behavior.discrete:
            callback_value = int(callback_value)

        return callback_value

    def value_delta(self, value, previous_value):
        if self.input_type == Configuration.INPUT_ROTATION_LEFT:
            return abs(value)
        if self.input_type == Configuration.INPUT_ROTATION_RIGHT:
            return abs(value)
        if self.input_type == Configuration.INPUT_SET:
            return abs(value - previous_value)
        if self.input_type == Configuration.INPUT_PRESSED:
            return 1
        if self.input_type == Configuration.INPUT_RELEASED:
            return 1

    def mandatory(self, value):
        if (self.behavior.behavior == ComponentBehavior.TRIGGER and
                self.behavior.limits and value in self.behavior.limits):
            return True

    def redundant(self, value):
        if (self.behavior.behavior == ComponentBehavior.CYCLE and
                value == self.persistent_state['stored_value']):
            return True

        if (self.behavior.behavior == ComponentBehavior.ACCUMULATE and
                value == self.persistent_state['stored_value']):
            return True

    def run(self, value, previous_value):
        change = self.value_delta(value, previous_value)

        self.cumulative_change += change

        if FLAGS.debug:
            print('[0] Evaluating callback %s %s (%s/%s) value %s change %s' % (
                self.input_type, self.action_name,
                self.cumulative_change, self.threshold, value, change))

        callback_value = self.callback_value(value)

        if not self.mandatory(callback_value) and self.cumulative_change < self.threshold:
            return

        if self.redundant(callback_value):
            return

        print('[0] Triggered %s callback %s %s callback_value %s' % (
            self.behavior.behavior, self.input_type, self.action_name, callback_value))

        self.cumulative_change = 0
        self.persistent_state['stored_value'] = callback_value

        if self.behavior.nested:
            callback_value = (callback_value, self.behavior.items[callback_value])

        self.function(callback_value, *self.data)

        return callback_value


class ComponentBehavior(object):
    CYCLE = 'CYCLE'
    TRIGGER = 'TRIGGER'
    ACCUMULATE = 'ACCUMULATE'
    DELTA = 'DELTA'

    behavior: int
    items: Any
    limits: Tuple[int, int]
    loop: bool
    invert: bool
    nested: bool
    discrete: bool
    threshold: int
    value: Any
    default: Any
    data: Any
    action_config: Dict[Any, Any]

    def __init__(self, behavior=None, items=None, limits=None,
            loop=None, invert=None, nested=None, discrete=None,
            threshold=None, value=None, default=None, action_config=None,
            data=None):
        self.behavior = behavior
        self.items = items
        self.limits = limits
        self.loop = loop
        self.invert = invert
        self.nested = nested
        self.discrete = discrete
        self.threshold = threshold
        self.value = value
        self.default = default
        self.data = data
        self.action_config = action_config

    def finalize(self):
        if not self.data:
            self.data = ()
        if not self.items:
            self.set_items(())
        if not self.default:
            self.default = 0
        if self.discrete is None:
            self.discrete = True
        if not self.action_config:
            self.action_config = {}
        if not self.threshold:
            self.threshold = 1
        self.set_items(self.items)
        self.merge_config()

    def merge_config(self):
        if 'value' in self.action_config:
            self.value = self.action_config['value']
        if 'invert' in self.action_config:
            self.invert = self.action_config['invert']
        if 'loop' in self.action_config:
            self.loop = self.action_config['loop']
        if 'default' in self.action_config:
            self.default = self.action_config['default']
        if 'discrete' in self.action_config:
            self.discrete = self.action_config['discrete']
        if 'items' in self.action_config:
            self.set_items(self.action_config['items'])
        if 'limits' in self.action_config:
            self.limits = self.action_config['limits']
        if 'threshold' in self.action_config:
            self.threshold = self.action_config['threshold']

    def __repr__(self):
        return '<ComponentBehavior %s items=%s limits=%s loop=%s invert=%s threshold=%s>' % (
                self.behavior, self.items, self.limits,
                self.loop, self.invert, self.threshold)

    def __add__(self, other):
        ret = copy.deepcopy(self)
        if other.behavior is not None:
            ret.behavior = other.behavior
        if other.limits is not None:
            ret.limits = other.limits
        if other.items is not None:
            ret.set_items(self.items + other.items)
        if other.loop is not None:
            ret.loop = other.loop
        if other.invert is not None:
            ret.invert = other.invert
        if other.nested is not None:
            ret.nested = other.nested
        if other.discrete is not None:
            ret.discrete = other.discrete
        if other.threshold is not None:
            ret.threshold = other.threshold
        if other.value is not None:
            ret.value = other.value
        if other.default is not None:
            ret.default = other.default
        if other.data is not None:
            ret.data = other.data
        if other.action_config is not None:
            ret.action_config = other.action_config
        ret.finalize()
        return ret

    def set_items(self, items):
        self.items = items
        if len(items):
            self.limits = (0, len(items) - 1)

    def outside(self, value):
        if not self.limits:
            return False
        if self.limits[0] is not None and value < self.limits[0]:
            return True
        if self.limits[1] is not None and value > self.limits[1]:
            return True
        return False


# FIXME: duck pattern probably makes sense for this
class Component(object):
    KIND_BASE = 0
    KIND_BUTTON = 1
    KIND_DIAL = 2
    KIND_SLIDER = 3

    KIND_NAME_MAP = {
        KIND_BASE: "Base",
        KIND_BUTTON: "Button",
        KIND_DIAL: "Dial",
        KIND_SLIDER: "Slider",
    }

    def __init__(self, index, uid, kind, flip=False):
        self.index = index
        self.uid = uid
        self.kind = kind
        self.flip = flip
        self.persistent_state = {}

        self.screen_string = ''
        self.screen_display = Configuration.ICON_BLANK

        self.values = [0, 0, 0, 0, 0, 0, 0, 0]
        self.previous_values = [0, 0, 0, 0, 0, 0, 0, 0]

        self.callbacks = None
        self.hidkeys = None
        self.num_ports = None

        # FIXME: do these belong in ComponentLayout?
        self.children = {}
        self.parent = None
        self.parent_port = None
        self.slot = None
        self.orientation = None
        self.led = (255, 255, 255)

        self.reset()

        self.set_sensitivity(9)

    def __repr__(self):
        return '<%s index=%d uid="%s" value="%s" orientation=%d flip=%s>' % (
                self.kind_name(), self.index, self.uid, self.values, self.orientation, self.flip)

    def reset(self):
        # TODO: should we also reset persistent_state here?
        self.callbacks = {}
        self.hidkeys = []

    def set_sensitivity(self, sensitivity):
        self.sensitivity = sensitivity

    def dominant_value(self, input_type):
        vindex = self.determine_vindex(input_type)
        return self.values[vindex]

    def determine_vindex(self, input_type):
        if input_type == Configuration.INPUT_PRESSED:
            return 0
        if input_type == Configuration.INPUT_RELEASED:
            return 0
        if self.kind == Component.KIND_DIAL:
            if input_type == Configuration.INPUT_SET:
                return 3
            if input_type == Configuration.INPUT_ROTATION_LEFT:
                return 1
            if input_type == Configuration.INPUT_ROTATION_RIGHT:
                return 2
        if self.kind == Component.KIND_SLIDER:
            return 0
        return 0

    def determine_input_types(self):
        input_types = set()

        if self.value_changed(0) and self.kind != Component.KIND_SLIDER:
            if self.values[0] == 1:
                input_types.add(Configuration.INPUT_PRESSED)
            if self.values[0] == 0:
                input_types.add(Configuration.INPUT_RELEASED)

        if self.kind == Component.KIND_DIAL:
            if self.value_changed(3):
                input_types.add(Configuration.INPUT_SET)
            if self.values[2] > 0:
                input_types.add(Configuration.INPUT_ROTATION_RIGHT)
            if self.values[1] > 0:
                input_types.add(Configuration.INPUT_ROTATION_LEFT)

        if self.kind == Component.KIND_SLIDER:
            input_types.add(Configuration.INPUT_SET)

        return list(input_types)

    def set_callback(self, input_type, action_name, behavior, callback, action_config):
        if behavior is None:
            return
        vindex = self.determine_vindex(input_type)
        print('[0] Bound callback %s %s (%s) to component %s' % (input_type, action_name, behavior.behavior, self.uid))
        invocation = ComponentInvocation(
                input_type, action_name, behavior,
                callback, vindex, action_config, self.persistent_state)
        self.callbacks[input_type] = invocation

    def invoke_callback(self, input_type):
        if input_type not in self.callbacks:
            return
        invocation = self.callbacks[input_type]
        vindex = self.determine_vindex(input_type)
        return invocation.run(self.values[vindex], self.previous_values[vindex])

    def set_hidkey(self, control, key, mod, repeat):
        self.hidkeys.append((control, key, mod, repeat))

    def set_orientation(self, orientation):
        self.orientation = orientation
        self.flip = False
        if self.kind == Component.KIND_SLIDER and orientation in (90, 180):
            self.flip = True

    def kind_name(self):
        return Component.KIND_NAME_MAP[self.kind]

    def value_changed(self, vindex):
        if len(self.previous_values) < vindex + 1:
            return True
        return self.previous_values[vindex] != self.values[vindex]


class ComponentLayout(object):
    components: Dict[str, Component]
    components_index: Dict[int, Component]
    components_slot: Dict[Tuple[int, int], Component]

    def __init__(self, layout_dict=None, orientation=0):
        self.base_orientation = orientation
        self.components = {}
        self.components_index = {}
        self.components_slot = {}
        self.layout_dict = {}
        if layout_dict:
            self.update(layout_dict)

    def __len__(self):
        return len(self.components)

    def update(self, layout_dict):
        self.components_index = {}
        self.components_slot = {}
        self.traverse(layout_dict)
        self.layout_dict = layout_dict

    def traverse(self, layout_dict, parent=None, port=None):
        if not layout_dict:
            return None

        # FIXME: the leading "  " in uid may be significant.
        # Maybe these are meant to be parsed into a number?
        uid = layout_dict["u"].lstrip()
        kind = layout_dict["t"]
        index = layout_dict["i"]
        conns = layout_dict["c"]

        orientation = self.calculate_orientation(parent, port)
        slot = self.calculate_slot(parent, port)

        component = self.components.get(uid,
                Component(index, uid, kind))
        component.index = index

        # TODO: consider always creating a new Component
        #component.reset()
        component.set_orientation(orientation)
        component.num_ports = len(conns)

        # FIXME: move these to be parameters on ComponentLayout
        component.parent = parent
        component.parent_port = port
        component.slot = slot
        component.children = {}

        for port, c in enumerate(conns):
            component.children[port] = self.traverse(c, component, port)

        self.components[uid] = component
        self.components_index[index] = uid
        self.components_slot[slot] = uid

        return component

    def calculate_slot(self, parent, port):
        if not parent:
            return (0,0)
        slot_calculator = ambit.coordinates.calculate_female_slots
        if parent.num_ports == 5:
            slot_calculator = ambit.coordinates.calculate_female_slots_wide
        return slot_calculator(parent.slot, parent.orientation)[port]

    def calculate_orientation(self, parent, port):
        if port is None:
            return self.base_orientation

        # this is arcane magic.
        # on a double-width component (5 female ports),
        # ports 2 and 3 face the same orientation as port 1 on square
        # port 4 has same orientation as port 2 on square
        # port 1 has same orientation as port 0 on square
        # port 0 has no equivalent but we need it to rotate -90 degrees
        if parent.num_ports == 5:
            if port == 0:
                port = -1
            else:
               port = int(port / 2.0)

        orientation = (parent.orientation - 90) + (port * 90)

        return orientation % 360

    def rowwise_index(component):
        pass

    def connected(self):
        components = []
        for index in sorted(self.components_index):
            uid = self.components_index[index]
            components.append(self.components[uid])
        return components

    def connected_rowwise(self):
        return sorted_rowwise(self.connected())

    # TODO: this does not need to be a class method
    def parse_query(self, query):
        if not (query.startswith('[') and query.endswith(']')):
            return {'uid': query}, None

        search_kwargs = {}

        # Parse the matcher portion of the query.
        matcher, *filters = query.lstrip('[').rstrip(']').split(' | ')
        matcher_parts = matcher.split(' ')
        for mp in matcher_parts:
            k, _, v = mp.partition('=')
            if k == 'kind':
                for kind, kind_name in Component.KIND_NAME_MAP.items():
                    if v == kind_name:
                        v = kind
                        break
            if k == 'index':
                v = int(v)
            if k == 'orientation':
                v = int(v)
            if k == 'slot':
                x, y = v.lstrip('(').rstrip(')').split(',')
                v = (int(x), int(y))
            search_kwargs[k] = v

        # Parse the filters.
        select = None
        for f in filters:
            if f == 'rowwise':
                search_kwargs['rowwise'] = True
            if f.startswith('select('):
                select = int(f.lstrip('select(').rstrip(')'))

        return search_kwargs, select

    def query(self, query):
        search_kwargs, select = self.parse_query(query)
        components = self.search(**search_kwargs)
        if select is not None:
            if len(components) < select:
                # FIXME: need better error handling for this case.
                print('[!] Query contains select with invalid index.')
                return []
            components = [components[select - 1]]
        #print('[D] Query %s matched components: %s' % (query, components))
        return components

    def search(self, uid=None, kind=None, index=None, slot=None, orientation=None, rowwise=False):
        components = []
        for component in self.connected():
            if uid is not None and component.uid != uid:
                continue
            if kind is not None and component.kind != kind:
                continue
            if slot is not None and component.slot != slot:
                continue
            if index is not None and component.index != index:
                continue
            if orientation is not None and component.orientation != orientation:
                continue
            components.append(component)
        if rowwise:
            components = sorted_rowwise(components)
        return components

    def find_component(self, index):
        if index not in self.components_index:
            return None
        uid = self.components_index[index]
        if uid not in self.components:
            return None
        return self.components[uid]

    def extrema(self):
        min_x, min_y = max_x, max_y = (0,0)
        for component in self.connected():
            x, y = component.slot
            # TODO: move this into component.dimensions?
            # in general, consolidate the uses of
            # component.orientation to fewer places.
            if component.num_ports == 5:
                if component.orientation == 0:
                    x += 1
                if component.orientation == 90:
                    y -= 1
                if component.orientation == 180:
                    x -= 1
                if component.orientation == 270:
                    y += 1
            if x < min_x:
                min_x = x
            if y < min_y:
                min_y = y
            if x > max_x:
                max_x = x
            if y > max_y:
                max_y = y
        return min_x, min_y, max_x, max_y

    def component_in_slot(self, slot):
        if slot not in self.components_slot:
            return None
        uid = self.components_slot[slot]
        return self.components[uid]



import argparse

flags = argparse.ArgumentParser(description='Take control of your Palette.')

flags.add_argument('--device', default='16D0:09F8',
                   help='usb device vendor:product id')

flags.add_argument('--device_index', default=0,
                   help='index of device to control if multiple connected')

flags.add_argument('--debug', default=False, action='store_true',
                   help='include low level protocol messages')

flags.add_argument('--verbose', default=False, action='store_true',
                   help='include more detailed status output')

flags.add_argument('--layouts', default=[], type=str, nargs='*',
                   help='launch with the specified layouts')

flags.add_argument('--map_midi', default=True, action='store_true',
                   help='map midi controller inputs')

flags.add_argument('config_paths', type=str, metavar='CONFIG', nargs='*',
                   help='configuration files to load')

FLAGS = flags.parse_args()



import ambit
import ambit.fake
import ambit.render
import pygame


class Controller(ambit.Controller):
    def __init__(self, config, device=None, simulate_input=True, pygame_event_callback=None):
        if device:
            self.device = device
        else:
            self.device = ambit.fake.Device('DEAD:BEEF', 'XYZ')

        super(Controller, self).__init__(config, self.device)

        self.device.components_connected()

        pygame_event_callbacks = []
        if pygame_event_callback:
            pygame_event_callbacks.append(pygame_event_callback)
        if simulate_input:
            pygame_event_callbacks.append(self.default_pygame_event_callback)

        self.display = ambit.render.Display(self,
                shutdown_event=self.shutdown_event,
                pygame_event_callbacks=pygame_event_callbacks)

    def default_pygame_event_callback(self, event):
        index = self.display.selected_component
        if not index:
            return
        if event.type == pygame.KEYDOWN and event.key == pygame.K_UP:
            self.device.input_slide_up(index)
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_DOWN:
            self.device.input_slide_down(index)
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_LEFT:
            self.device.input_rotation_left(index)
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_RIGHT:
            self.device.input_rotation_right(index)
        elif event.type == pygame.KEYDOWN and event.key == pygame.K_SPACE:
            self.device.input_pressed(index)
        elif event.type == pygame.KEYUP and event.key == pygame.K_SPACE:
            self.device.input_released(index)

    def connect(self):
        super(Controller, self).connect()
        self.display.run()



# TODO: replace these with cartesian rotation as defined here:
# https://en.wikipedia.org/wiki/Cartesian_coordinate_system#Cartesian_formulae_for_the_plane


def calculate_male_slot(slot, orientation):
    x, y = slot
    if orientation == 0:
        return (x, y + 1)
    if orientation == 90:
        return (x + 1, y)
    if orientation == 180:
        return (x, y - 1)
    if orientation == 270:
        return (x - 1, y)


def calculate_female_slots(slot, orientation):
    x, y = slot
    if orientation == 0:
        return (x + 1, y), (x, y - 1), (x - 1, y)
    if orientation == 180:
        return (x - 1, y), (x, y + 1), (x + 1, y)
    if orientation == 90:
        return (x, y - 1), (x - 1, y), (x, y + 1)
    if orientation == 270:
        return (x, y + 1), (x + 1, y), (x, y - 1)


def calculate_female_slots_wide(slot, orientation):
    x, y = slot
    if orientation == 0:
        return (x + 1, y + 1), (x + 2, y), (x + 1, y - 1), (x, y - 1), (x - 1, y)
    if orientation == 180:
        return (x - 1, y - 1), (x - 2, y), (x - 1, x + 1), (x, y + 1), (x + 1, y)
    if orientation == 90:
        return (x + 1, y - 1), (x, y - 2), (x - 1, y - 1), (x - 1, y), (x, y + 1)
    if orientation == 270:
        return (x - 1, y + 1), (x, y + 2), (x + 1, y + 1), (x + 1, y), (x, y - 1)



import glob
import os
import pathlib
import setuptools

package_data = dict()
package_data['ambit'] = []
for path, paths, files in os.walk('ambit/resources/', followlinks=True):
    for f in files:
        include = os.path.join(path, f)[6:]
        package_data['ambit'].append(include)

setuptools.setup(
    name='ambit',
    version='0.3.6',
    description='Take control of your Palette.',
    long_description=(
        '**ambit** is a Python library for interacting with PaletteGear '
        'and MonogramCC devices, a graphical simulator for device-free '
        'development, and an accompanying set of configurable end user '
        'tools and demos.'
    ),
    long_description_content_type='text/markdown',
    keywords=['monogramcc', 'palettegear', 'demoscene', 'pygame', 'linux', 'pyusb'],
    url='https://github.com/khimaros/ambit',
    author='khimaros',
    packages=setuptools.find_packages(),
    package_data=package_data,
    scripts=glob.glob('./bin/*'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX',
        'Topic :: System :: Hardware',
    ],
    python_requires='>=3.6',
)




import ambit.resources

from ambit.component import Component, ComponentBehavior, ComponentLayout
from ambit.configuration import Configuration
from ambit.message import message_decode, message_encode
from ambit.flags import FLAGS

import os
import subprocess
import sys
import threading
import time
import usb
import queue

from typing import Any, Callable, Dict, List, Tuple


class Device:
    def __init__(self, device_id, interface_id, index=0):
        vendor_id, product_id = device_id.split(':')
        self.vendor_id = int(vendor_id, 16)
        self.product_id = int(product_id, 16)
        self.interface_id = interface_id
        self.index = index
        self._device = self._discover()
        self.handle = None

    def _discover(self):
        index = 0
        buses = usb.busses()
        for bus in buses:
            for device in bus.devices:
                if device.idVendor == self.vendor_id:
                    if device.idProduct == self.product_id:
                        if index == int(self.index):
                            print('[0] Choosing device at USB discovery index %d.' % index)
                            return device
                        index += 1
        return None

    def open(self):
        if not self._device:
            print('[!] Failed to read from device, check cable and permissions!')
            sys.exit(1)
        handle = self._device.open()
        try:
            handle.detachKernelDriver(self.interface_id)
        except:
            pass
        handle.claimInterface(self.interface_id)
        self.handle = handle
        return handle


class Controller:
    BULK_WRITE_TIMEOUT_MS = 1000
    BULK_READ_TIMEOUT_MS = 1000
    BULK_READ_SIZE_BYTES = 4096
    QUEUE_TIMEOUT_SECONDS = 0.5

    DEFAULT_WAIT_SECONDS = 0.5

    # Above 20 updates per second, the screen will hard reset
    # if the string length is larger than ~3 bytes.
    SCREEN_STRING_DELAY_SECONDS = 1 / 20
    SCREEN_STRING_QUEUE_DEPTH = 8
    LED_DELAY_SECONDS = 1 / 60
    LED_QUEUE_DEPTH = 64

    KEEPALIVE_TIMEOUT_SECONDS = 5
    SCREEN_RESET_SECONDS = 3

    USB_INTERFACE_ID = 0                # Interface for CDC endpoint.
    USB_ENDPOINT_CONTROL_IN = 0x80      # CDC Control - EP 0 IN
    USB_ENDPOINT_CONTROL_OUT = 0x00     # CDC Control - EP 0 OUT
    USB_ENDPOINT_INTERRUPT_IN = 0x83    # CDC Interrupt - EP 3 IN
    USB_ENDPOINT_BULK_IN = 0x84         # CDC Data (Bulk) - EP 4 IN
    USB_ENDPOINT_BULK_OUT = 0x05        # CDC Data (Bulk) - EP 5 OUT
    USB_REQUEST_TYPE_IN = 0xa1          # CDC Control - Request 161
    USB_REQUEST_TYPE_OUT = 0x21         # CDC Control - Request 33
    USB_SETUP_REQUEST_HANDSHAKE = 0x20  # 32
    USB_SETUP_REQUEST_EMPTY_IN = 0x21   # 33
    USB_SETUP_REQUEST_EMPTY_OUT = 0x22  # 34

    # Magic string used during control handshake.
    HANDSHAKE_BYTES = [0xe0, 0x93, 0x04, 0x00, 0x00, 0x00, 0x00, 0x08]

    FIRMWARE_VERSION_SLIDER_FIXED = '1.3.2'

    BEHAVIOR_KIND_DEFAULTS = {
            Component.KIND_SLIDER: ComponentBehavior(
                behavior=ComponentBehavior.TRIGGER,
                limits=(0,255), invert=False),
    }
    BEHAVIOR_INPUT_DEFAULTS = {
            Configuration.INPUT_PRESSED: ComponentBehavior(
                behavior=ComponentBehavior.TRIGGER),
            Configuration.INPUT_RELEASED: ComponentBehavior(
                behavior=ComponentBehavior.TRIGGER),
            Configuration.INPUT_ROTATION_LEFT: ComponentBehavior(
                behavior=ComponentBehavior.ACCUMULATE, invert=True),
            Configuration.INPUT_ROTATION_RIGHT: ComponentBehavior(
                behavior=ComponentBehavior.ACCUMULATE, invert=False),
            Configuration.INPUT_SET: ComponentBehavior(
                behavior=ComponentBehavior.TRIGGER, invert=False, limits=(0,255)),
    }
    BEHAVIOR_ACTION_DEFAULTS = {
            Configuration.ACTION_SET_COLOR_RED: ComponentBehavior(
                limits=(0, 255), default=255, threshold=8),
            Configuration.ACTION_SET_COLOR_GREEN: ComponentBehavior(
                limits=(0, 255), default=255, threshold=8),
            Configuration.ACTION_SET_COLOR_BLUE: ComponentBehavior(
                limits=(0, 255), default=255, threshold=8),

            Configuration.ACTION_CYCLE_MAPPING: ComponentBehavior(
                ComponentBehavior.CYCLE, loop=True, nested=True),

            Configuration.ACTION_PROFILE_NEXT: ComponentBehavior(
                ComponentBehavior.TRIGGER),
            Configuration.ACTION_PROFILE_PREV: ComponentBehavior(
                ComponentBehavior.TRIGGER),
            Configuration.ACTION_PROFILE_SWITCH: ComponentBehavior(
                ComponentBehavior.CYCLE, loop=True),

            Configuration.ACTION_TEST_CYCLE: ComponentBehavior(
                ComponentBehavior.CYCLE, loop=True, nested=True),
            Configuration.ACTION_TEST_DELTA: ComponentBehavior(
                ComponentBehavior.DELTA),
            Configuration.ACTION_TEST_ACCUMULATE: ComponentBehavior(
                ComponentBehavior.ACCUMULATE),
            Configuration.ACTION_TEST_TRIGGER: ComponentBehavior(
                ComponentBehavior.TRIGGER),
    }

    device: Device
    handle: usb.legacy.DeviceHandle
    config: Configuration
    layout: ComponentLayout
    lock: threading.Lock
    shutdown_event: threading.Event
    write_queue: queue.Queue
    action_callback_map: Dict[str, Callable]
    version_core: str

    def __init__(self, config=None, device=None):
        self.device = device or Device(FLAGS.device, Controller.USB_INTERFACE_ID, FLAGS.device_index)
        self.handle = None
        self.config = config
        self.layout = ComponentLayout(orientation=config.profile.orientation)

        # TODO: do we need more than one lock?
        self.lock = threading.Lock()
        self.shutdown_event = threading.Event()

        self.write_queue = queue.Queue()
        self.screen_string_queue = queue.Queue(Controller.SCREEN_STRING_QUEUE_DEPTH)
        self.led_queue = queue.Queue(Controller.LED_QUEUE_DEPTH)

        self.bulk_read_thread = threading.Thread(target=self.bulk_read_worker)
        self.bulk_write_thread = threading.Thread(target=self.bulk_write_worker)
        self.screen_string_thread = threading.Thread(target=self.screen_string_worker)
        self.led_thread = threading.Thread(target=self.led_worker)
        self.keepalive_thread = threading.Thread(target=self.keepalive_worker)

        self.version_core = ''
        self.failed_writes = 0
        self.failed_reads = 0
        self.dropped_screen_strings = 0
        self.dropped_leds = 0

        self.last_led_time = 0
        self.last_write_time = 0
        self.last_screen_string_time = 0

        # FIXME: this doesn't feel like the right home
        self.current_led_values = (255, 255, 255)
        self.current_cycle_mappings = {}
        self.current_screen_string_value = ''

        default_action_callbacks = {
            Configuration.ACTION_PROFILE_PREV: self.callback_profile_prev,
            Configuration.ACTION_PROFILE_NEXT: self.callback_profile_next,
            Configuration.ACTION_PROFILE_SWITCH: self.callback_profile_switch,
            Configuration.ACTION_SET_COLOR_RED: self.callback_set_color_red,
            Configuration.ACTION_SET_COLOR_GREEN: self.callback_set_color_green,
            Configuration.ACTION_SET_COLOR_BLUE: self.callback_set_color_blue,
            Configuration.ACTION_EXECUTE_COMMAND: self.callback_execute_command,
            Configuration.ACTION_CYCLE_MAPPING: self.callback_cycle_mapping,
            Configuration.ACTION_TOGGLE_PREVIEW: lambda x: None,
            Configuration.ACTION_TEST_CYCLE: self.callback_test_cycle,
            Configuration.ACTION_TEST_DELTA: self.callback_test_delta,
            Configuration.ACTION_TEST_ACCUMULATE: self.callback_test_accumulate,
            Configuration.ACTION_TEST_TRIGGER: self.callback_test_trigger,
        }

        self.action_callback_map = {}
        for action_name in default_action_callbacks:
            action_callback = default_action_callbacks[action_name]
            self.register_action_callback(action_name, action_callback)

    def register_action_callback(self, action_name, callback):
        self.action_callback_map[action_name] = callback

    def open(self):
        self.handle = self.device.open()

    def close(self):
        """ Release device interface """
        print('[0] Joining threads and releasing USB interface.')
        self.join()
        try:
            self.handle.reset()
            self.handle.releaseInterface()
            #self.handle.attachKernelDriver(self.device.interface_id)
        except Exception as err:
            print(err)
        self.handle, self.device = None, None

    # TODO: turn these hex values into descriptive constants
    def control_handshake(self):
        # URB_CONTROL in (empty)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_IN,
                Controller.USB_SETUP_REQUEST_EMPTY_IN, 7)

        # URB_CONTROL out (handshake)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_OUT,
                Controller.USB_SETUP_REQUEST_HANDSHAKE,
                Controller.HANDSHAKE_BYTES)

        # URB_CONTROL in (empty)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_IN,
                Controller.USB_SETUP_REQUEST_EMPTY_IN, 7)

        # URB_CONTROL out (empty)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_OUT,
                Controller.USB_SETUP_REQUEST_EMPTY_OUT, 0)

        # URB_CONTROL out (handshake)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_OUT,
                Controller.USB_SETUP_REQUEST_HANDSHAKE,
                Controller.HANDSHAKE_BYTES)

        # URB_CONTROL in (empty)
        self.handle.controlMsg(
                Controller.USB_REQUEST_TYPE_IN,
                Controller.USB_SETUP_REQUEST_EMPTY_IN, 7)

    def bulk_write_worker(self):
        while not self.shutdown_event.is_set():
            try:
                data = self.write_queue.get(timeout=Controller.QUEUE_TIMEOUT_SECONDS)
            except queue.Empty:
                pass
            else:
                self.bulk_write(data)
                self.write_queue.task_done()

    def bulk_write_messages(self, messages):
        if not messages:
            return
        data = message_encode(messages)
        self.write_queue.put(data, Controller.QUEUE_TIMEOUT_SECONDS)

    def bulk_write(self, data):
        if FLAGS.debug:
            print("[@] BULK WRITE:", data.tobytes().decode())
        try:
            self.handle.bulkWrite(
                    Controller.USB_ENDPOINT_BULK_OUT,
                    data,
                    Controller.BULK_WRITE_TIMEOUT_MS)
        except usb.USBError as exc:
            self.failed_writes += 1
            self.process_usb_error(exc)
        else:
            self.lock.acquire()
            self.last_write_time = time.time()
            self.lock.release()

    def bulk_read_worker(self):
        while not self.shutdown_event.is_set():
            messages = self.bulk_read()
            for message in messages:
                self.process_bulk_message(message)

    def bulk_read(self):
        try:
            data = self.handle.bulkRead(
                    Controller.USB_ENDPOINT_BULK_IN,
                    Controller.BULK_READ_SIZE_BYTES,
                    Controller.BULK_READ_TIMEOUT_MS)
        except usb.USBError as exc:
            self.failed_reads += 1
            self.process_usb_error(exc)
            return []
        else:
            if FLAGS.debug:
                print("[@] BULK READ:", data.tobytes().decode())
            messages, _ = message_decode(data)
            return messages

    def process_usb_error(self, exc):
        if exc.errno == 19:
            print('[0] Lost connection to device, shutting down.')
            self.join()

    def start(self):
        self.bulk_write_messages([
            {"start": 1},
        ])

    def stop(self):
        self.bulk_write_messages([
            {"stop": 1},
        ])

    def check(self):
        self.bulk_write_messages([
            {"check": 1},
        ])

    def led_worker(self):
        while not self.shutdown_event.is_set():
            if self.led_queue.full():
                # TODO: should we be clearing the entire queue when full, or just some of it?
                for _ in range(self.led_queue.qsize() - 2):
                    self.led_queue.get()
                    self.led_queue.task_done()
                    self.dropped_leds += 1
            if time.time() - self.last_led_time < Controller.LED_DELAY_SECONDS:
                wait = Controller.LED_DELAY_SECONDS - (time.time() - self.last_led_time)
                time.sleep(wait)
            try:
                # TODO: look ahead in the queue to find duplicates or
                # ways that we can join multiple updates into a single
                # bulk write?
                leds = self.led_queue.get(timeout=Controller.QUEUE_TIMEOUT_SECONDS)
            except queue.Empty:
                pass
            else:
                data = message_encode([
                    {"led": leds},
                ])
                self.bulk_write(data)
                # FIXME: this feels like a pretty inefficient way to
                # update the component state, but fixing may require
                # changing the API surface provided by Controller.led()
                for led in leds:
                    component = self.layout.find_component(led['i'])
                    if component:
                        component.led = (led['r'], led['g'], led['b'])
                self.last_led_time = time.time()
                self.led_queue.task_done()

    def led(self, leds):
        try:
            self.led_queue.put(leds, timeout=Controller.QUEUE_TIMEOUT_SECONDS)
        except queue.Full:
            self.dropped_leds += 1

    def clear(self):
        # Reset mappings to empty
        self.bulk_write_messages([
            { "hidmap": [] },
            { "joymap": [] },
            { "midimap": [] },
            { "led": [] },
        ])

    def cycle(self):
        # Cycle through each LED and reset to white.
        for component in self.layout.connected():
            self.led([
                {"b": 0, "g": 0, "i": component.index, "m": 0, "r": 255},
            ])
            self.led([
                {"b": 255, "g": 255, "i": component.index, "m": 0, "r": 255},
            ])
            cr, cg, cb = self.current_led_values
            self.led([
                {"b": cb, "g": cg, "i": component.index, "m": 0, "r": cr},
            ])

    def print_stats(self):
        print('[@] Cumulative dropped_leds:', self.dropped_leds)
        print('[@] Cumulative dropped_screen_strings:', self.dropped_screen_strings)
        print('[@] Cumulative failed_writes:', self.failed_writes)

    def screen_string_worker(self):
        while not self.shutdown_event.is_set():
            if time.time() - self.last_screen_string_time < Controller.SCREEN_STRING_DELAY_SECONDS:
                wait = Controller.SCREEN_STRING_DELAY_SECONDS - (time.time() - self.last_screen_string_time)
                time.sleep(wait)
            if self.screen_string_queue.full():
                # TODO: should we be clearing the entire queue when full, or just some of it?
                for _ in range(self.screen_string_queue.qsize() - 2):
                    self.screen_string_queue.get(timeout=Controller.QUEUE_TIMEOUT_SECONDS)
                    self.screen_string_queue.task_done()
                    self.dropped_screen_strings += 1
            try:
                title = self.screen_string_queue.get(timeout=Controller.QUEUE_TIMEOUT_SECONDS)
            except queue.Empty:
                pass
            else:
                data = message_encode([
                    { "screen_string": str(title) },
                ])
                self.bulk_write(data)
                component = self.layout.find_component(1)
                component.screen_string = str(title)
                self.last_screen_string_time = time.time()
                self.current_screen_string_value = title
                self.screen_string_queue.task_done()

    def screen_string(self, title):
        if title == self.current_screen_string_value:
            return
        try:
            self.screen_string_queue.put(title, timeout=Controller.QUEUE_TIMEOUT_SECONDS)
        except queue.Full:
            self.dropped_screen_strings += 1

    def screen_write(self, path, index):
        print('[0] Uploading image %s to index %d' % (path, index))
        with open(path, 'rb') as f:
            data = message_encode([{'screen_write': index}], f.read())
            flen = f.tell()
        self.handle.bulkWrite(Controller.USB_ENDPOINT_BULK_OUT, data, 10000)
        print('[0] Image upload to index %d (%d bytes) complete!'  % (index, flen))

    def screen_display(self, index):
        self.bulk_write_messages([
            {"screen_display": index},
        ])
        component = self.layout.find_component(1)
        component.screen_display = index

    def initialize(self):
        # TODO: initialize layout using configuration and set up
        # callbacks before the first layout update.
        #self.layout.update(self.config.profile.layout)
        #self.configure_component_callbacks()

        self.start()

        self.clear()

        self.send_version()

        print("[0] Waiting for initial layout ...")
        self.wait_for_layout()

        self.check()

    def configure_orientation(self):
        messages = []
        for component in self.layout.connected():
            flip = 1 if component.flip else 0
            messages.append({"flip": [ {"i": component.index, "m": flip } ]})
        self.bulk_write_messages(messages)

    def send_version(self):
        self.bulk_write_messages([
            {"send_version": 1},
        ])

    def boot(self):
        if len(self.layout.connected()) > 1:
            print('[0] Skipping boot, too many components connected.')
            return

        self.check()

        self.screen_display(0)

        self.screen_string("Bootloader")

        time.sleep(1)

        self.bulk_write_messages([ { "boot": 1 } ])

        print('[0] Rebooting to bootloader...')

        self.wait()

        time.sleep(1)

    def configure_images(self):
        if len(self.layout.connected()) > 1:
            print('[0] Skipping image upload, too many components connected.')
            return

        self.check()
        self.clear()

        #for index in range(25):
        #    asset_path = ambit.resources.asset_path(index)
        #    self.screen_write(asset_path, index)
        #    self.screen_display(index)
        #    time.sleep(1)

        for index in (23, 24, 25):
            asset_path = ambit.resources.asset_path(index)
            self.screen_write(asset_path, index)
            self.screen_display(index)
            time.sleep(1)

        self.bulk_write_messages([ { "set_version_screen": 0 } ])

        self.start()

    def led_draw(self, pixels, x, y, w, h):
        leds = []
        for cy in range(y, y - h, -1):
            for cx in range(x, x + w):
                component = self.layout.component_in_slot((cx, cy))
                if not component:
                    continue
                r, g, b = pixels.pop(0)
                leds.append({"b": b, "g": g, "i": component.index, "m": 0, "r": r})
        self.led(leds)

    def configure_leds(self, red=None, green=None, blue=None):
        if red is None: red = self.current_led_values[0]
        if green is None: green = self.current_led_values[1]
        if blue is None: blue = self.current_led_values[2]

        leds = []
        for component in self.layout.connected():
            leds.append({"b": blue, "g": green, "i": component.index, "m":0, "r": red})
        self.led(leds)

        # FIXME: this may not represent current state as self.led() is asynchronous
        self.current_led_values = (red, green, blue)

    def configure_midimap_dial(self, component):
        #rotate_note = component.index
        #pressed_note = component.index * 12
        rotation_note = self.component_hash(component)
        pressed_note = self.component_hash(component)
        self.bulk_write_messages([
            {"midimap":[{
                "i": component.index,
                "s": [
                    # dial dial rotation
                    {"c":3, "d": rotation_note, "m":1, "mc":0, "t":2},
                    # dial button press
                    {"c":0, "d": pressed_note, "m":0, "mc":0, "t":1}
                ],
            }]},
        ])

    def configure_midimap_slider(self, component):
        #note = component.index * 12
        note = self.component_hash(component)
        self.bulk_write_messages([
            {"midimap": [{"i": component.index, "s": [{ "c": 0, "d": note, "m" :0, "mc": 0, "t": 2 }]}]},
        ])

    def configure_midimap_button(self, component):
        #note = component.index * 12
        note = self.component_hash(component)
        self.bulk_write_messages([
            {"midimap": [{
                "i": component.index,
                "s": [
                    { "c": 0, "d": note, "m": 0, "mc": 0, "t": 1 },
                ]},
            ]},
        ])

    def component_hash(self, component):
        # possible component.uid are 32 through 126 (94 possible values)
        # with 2 characters, 94 * 94 = 8836 possible combinations.
        # but there are only 127 possible note values in MIDI, so we need
        # to find a way to compress.
        table = list(range(0, 128))
        h = len(component.uid[1:2]) % 128
        for i in component.uid[1:]:
            h = table[h ^ ord(i)]
        return h

    # FIXME: keep consistent mapping across disconnect/reconnect
    # need to remove reliance on index (use uid instead)
    def configure_midimap(self):
        for component in self.layout.connected():
            if component.kind == Component.KIND_DIAL:
                self.configure_midimap_dial(component)
            if component.kind == Component.KIND_BUTTON:
                self.configure_midimap_button(component)
            if component.kind == Component.KIND_SLIDER:
                self.configure_midimap_slider(component)

    def configure_hidmap(self):
        hidmaps = []
        for component in self.layout.connected():
            print('HIDKEYS:', component.hidkeys)
            for control, key, mod, repeat in component.hidkeys:
                hm = { "i": component.index, "s": [{
                    "c": control, "k": key, "m": mod, "r": repeat, "rp": repeat}], "t": 0 }
                self.bulk_write_messages([{ "hidmap": [hm] }])
                hidmaps.append(hm)
        #self.bulk_write_messages([{ "hidmap": hidmaps }])

    def configure_display(self):
        orientation = int((180 + self.config.profile.orientation) % 360 / 90)
        self.bulk_write_messages([
            { "screen_orientation": orientation },
            { "screen_display": self.config.profile.icon },
            { "screen_string": self.config.profile.title },
        ])
        component = self.layout.find_component(1)
        component.screen_string = self.config.profile.title
        component.screen_display = self.config.profile.icon

    def configure_component_hidkeys(self):
        for component in self.layout.connected():
            self.set_component_hidkeys(component)

    def set_component_hidkeys(self, component):
        for control, key, mod, repeat in self.config.component_hidkeys(component.uid):
            component.set_hidkey(control, key, mod, repeat)

    # TODO: this function could be moved to out of Controller along with the maps.
    # could be a static method eg. ComponentBehavior.from_action(...) and might
    # be possible to remove action_config param from ComponentBehavior.__init__()
    def make_action_behavior(self, action_config, component_kind, input_type, action_name):
        behavior = ComponentBehavior(action_config=action_config)
        if component_kind in Controller.BEHAVIOR_KIND_DEFAULTS:
            behavior += Controller.BEHAVIOR_KIND_DEFAULTS[component_kind]
        if input_type in Controller.BEHAVIOR_INPUT_DEFAULTS:
           behavior += Controller.BEHAVIOR_INPUT_DEFAULTS[input_type]
        if action_name in Controller.BEHAVIOR_ACTION_DEFAULTS:
            behavior += Controller.BEHAVIOR_ACTION_DEFAULTS[action_name]
        return behavior

    # TODO: Move this (and callback methods) to outside of Ambit class
    # and instead allow subclasses to set them up. We can have a common
    # subclass which is the default behavior.
    def configure_component_callbacks(self):
        for component in self.layout.connected():
            component.reset()
        for uid_query in self.config.components():
            components = self.layout.query(uid_query)
            if components:
                component = components[0]
                self.set_component_callbacks(uid_query, component)

    def set_component_callbacks(self, uid_query, component):
        for input_type, action_name, action_config in self.config.component_actions(uid_query):
            if action_name not in self.action_callback_map:
                print('[0] Unrecognized action type:', action_name)
                continue
            callback = self.action_callback_map[action_name]

            behavior = self.make_action_behavior(action_config, component.kind, input_type, action_name)
            if action_name == Configuration.ACTION_EXECUTE_COMMAND:
                behavior.data = [action_config['argv']]
            if action_name == Configuration.ACTION_CYCLE_MAPPING:
                behavior.data = [action_config['target']]
            if action_name == Configuration.ACTION_PROFILE_SWITCH:
                behavior.set_items(self.config.profiles)

            component.set_callback(input_type, action_name, behavior, callback, action_config)

    def callback_test_cycle(self, value):
        _, value = value
        self.screen_string('C: %s' % value)

    def callback_test_delta(self, value):
        self.screen_string('D: %s' % value)

    def callback_test_accumulate(self, value):
        self.screen_string('A: %s' % value)

    def callback_test_trigger(self, value):
        self.screen_string('T: %s' % value)

    def callback_cycle_mapping(self, value, target):
        index, action_config = value
        component = self.layout.query(target)[0]
        self.config.set_component_actions(component.uid, action_config)
        self.set_component_callbacks(component.uid, component)
        self.screen_string(['Red', 'Green', 'Blue'][index])

    def callback_execute_command(self, value, argv):
        cmd = []
        for a in argv:
            cmd.append(a.replace('%AMBIT_VALUE%', str(value)))
        print('[0] Executing command: %s' % cmd)
        p = subprocess.run(
                cmd, env=os.environ.update({'AMBIT_VALUE': str(value)}),
                stdout=subprocess.PIPE)
        result = p.stdout.decode().split('\n')[0].strip()
        if result:
            self.screen_string(result)

    def callback_set_color_red(self, value):
        value = int(value)
        self.configure_leds(red=value)
        self.screen_string('Red: %d' % value)

    def callback_set_color_green(self, value):
        value = int(value)
        self.configure_leds(green=value)
        self.screen_string('Green: %d' % value)

    def callback_set_color_blue(self, value):
        value = int(value)
        self.configure_leds(blue=value)
        self.screen_string('Blue: %d' % value)

    def callback_profile_switch(self, value):
        value = int(value)
        self.config.switch(value)
        self.start()

    def callback_profile_prev(self, unused_value):
        self.config.prev()
        self.start()

    def callback_profile_next(self, unused_value):
        self.config.next()
        self.start()

    def print_layout(self):
        for component in self.layout.connected():
            bound_inputs = []
            for input_type, action_name, _ in self.config.component_actions(component.uid):
                bound_inputs.append('%s (%s)' % (input_type, action_name))
            print('[0] Component attached: %s (%d) %s %s - %s - %s :: %s' % (
                component.kind_name(), component.index, component.uid,
                component.slot, component.orientation, component.flip,
                ', '.join(bound_inputs)))

    def print_input(self, component, input_type, value, raw_value):
        if not FLAGS.verbose:
            return
        print('[%d] Input %s %s: %s (%d)!' % (
            component.index, component.kind_name(),
            input_type, value, raw_value))

    def display_input(self, component, input_type, value, raw_value):
        if not FLAGS.verbose:
            return
        sv = str(value) if value is not None else str(raw_value)
        self.screen_string('%s: %s' % (component.kind_name(), sv))

    # FIXME: this function is reaching into all kinds of private areas
    def rotate(self, rotation):
        orientation = (self.layout.base_orientation + rotation) % 360
        self.config.profile.orientation = orientation
        self.layout.base_orientation = orientation
        self.process_layout(self.layout.layout_dict)

    def process_bulk_message(self, message):
        if 'l' in message:
            self.process_layout(message['l'])
        elif 'in' in message:
            self.process_input(message['in'])
        elif 'version_core' in message:
            self.process_version_core(message['version_core'])

    def process_layout(self, layout_dict):
        self.layout.update(layout_dict)

        self.print_layout()
        self.clear()
        self.cycle()

        self.configure_component_callbacks()
        self.configure_orientation()
        self.configure_leds()
        self.configure_component_hidkeys()
        self.configure_display()

        # TODO: support layout change callback
        #self.layout_changed_callback()

        if FLAGS.map_midi:
            self.configure_midimap()
        #self.configure_hidmap()

        print('[0] Processed layout, ready for input!')

    def process_input(self, input_messages):
        for input_dict in input_messages:
            i = input_dict["i"]
            v = input_dict["v"]
            component = self.layout.find_component(i)
            if not component:
                return

            # Workaround bug in firmware version 1.3.1 which causes slider
            # inputs to stop at 254 instead of 255.
            if self.version_core < Controller.FIRMWARE_VERSION_SLIDER_FIXED:
                if component.kind == Component.KIND_SLIDER:
                    if v[0] == 254:
                        v[0] = 255

            component.previous_values = component.values
            component.values = v
            for input_type in component.determine_input_types():
                raw_value = component.dominant_value(input_type)
                value = component.invoke_callback(input_type)
                self.print_input(component, input_type, value, raw_value)
                self.display_input(component, input_type, value, raw_value)

    def process_version_core(self, version):
        print('[0] Detected Palette firmware version', version)
        self.version_core = version

    def wait_for_layout(self):
        while not self.layout:
            time.sleep(Controller.DEFAULT_WAIT_SECONDS)

    def wait(self):
        while self.led_queue.qsize() > 0:
            time.sleep(Controller.DEFAULT_WAIT_SECONDS)
        while self.screen_string_queue.qsize() > 0:
            time.sleep(Controller.DEFAULT_WAIT_SECONDS)
        while self.write_queue.qsize() > 0:
            time.sleep(Controller.DEFAULT_WAIT_SECONDS)

    def join(self):
        self.shutdown_event.set()
        all_threads = (
                self.keepalive_thread,
                self.screen_string_thread,
                self.led_thread,
                self.bulk_write_thread,
                self.bulk_read_thread,
        )
        for thread in all_threads:
            if thread.is_alive():
                try:
                    thread.join()
                except:
                    pass

    def spawn(self):
        self.bulk_read_thread.start()
        self.bulk_write_thread.start()
        self.screen_string_thread.start()
        self.led_thread.start()

    def connect(self):
        self.bulk_read()
        self.control_handshake()
        self.spawn()
        self.initialize()
        self.keepalive_thread.start()

    def keepalive_worker(self):
        while not self.shutdown_event.is_set():
            cur_time = time.time()
            self.lock.acquire()
            last_write_time = self.last_write_time
            last_screen_string_time = self.last_screen_string_time
            screen_title = self.config.profile.title
            self.lock.release()
            if cur_time - last_screen_string_time > Controller.SCREEN_RESET_SECONDS:
                if self.config.profile.title is not None:
                    self.screen_string(screen_title)
            if cur_time - last_write_time > Controller.KEEPALIVE_TIMEOUT_SECONDS:
                self.check()
            time.sleep(Controller.DEFAULT_WAIT_SECONDS)

    def communicate(self):
        try:
            while not self.shutdown_event.is_set():
                time.sleep(Controller.DEFAULT_WAIT_SECONDS)
        except KeyboardInterrupt:
            print('\r[0] Received interrupt, shutting down...')
        except Exception as err:
            self.close()
            raise err


# TODO: make this a thing! (maybe subclass?)
class StandardAmbit:
    def __init__(self, config_path):
        pass



from ambit.flags import FLAGS

import json

from typing import Any, Callable, Dict, List, Tuple


def message_encode(messages: List[str], extra_data: bytes=bytes(())) -> memoryview:
    json_messages = []
    for message in messages:
        json_message = json.dumps(message, separators=(',', ':'))
        json_messages.append(json_message)
    data = "".join(json_messages)
    return memoryview(bytes(data, 'utf8') + extra_data)


def message_decode(data: memoryview) -> Tuple[List[str], bytes]:
    extra_data = bytes([])
    # Parse out data bytes from the rest of the message.
    data_bytes = data.tobytes()
    if data_bytes.startswith(b'{"screen_write":'):
        extra_data_start = data_bytes.index(b'}') + 1
        extra_data = data_bytes[extra_data_start:]
        data_bytes = data_bytes[:extra_data_start]

    try:
        read_str = data_bytes.decode()
    except:
        print('[!] Failed to decode bytes:', data_bytes);
        return [], extra_data

    # FIXME: this is an awful, terrible hack which will fail as soon
    # as it encounters }{ inside of a string.
    json_data = '[' + read_str.strip().replace('\n', ',').replace('}{', '},{') + ']'

    try:
        messages = json.loads(json_data)
    except json.decoder.JSONDecodeError as err:
        if FLAGS.debug:
            print('[!] DECODE FAILED:', json_data, "--", err)
        messages = []

    return messages, extra_data
